#include "VisMonitoring/DQMServer/interface/VisDQMImageServer.h"
#include "VisMonitoring/DQMServer/interface/DQMRenderPlugin.h"
#include "VisMonitoring/DQMServer/interface/DQMRenderPluginDB.h"
#include "Iguana/Framework/interface/IgPluginManager.h"
#include "classlib/utils/Regexp.h"
#include "classlib/utils/StringFormat.h"
#include "classlib/utils/SystemError.h"
#include "classlib/utils/Error.h"
#include "TObjString.h"
#include "TCanvas.h"
#include "TImageDump.h"
#include "TASImage.h"
#include "TText.h"
#include "TColor.h"
#include "TAxis.h"
#include "TList.h"
#include "TH1.h"
#include "TProfile.h"
#include "TROOT.h"
#include "TStyle.h"
#include <iostream>
#include <sys/time.h>
#include <math.h>

using namespace lat;

static const Regexp s_rxmeval("<(.*)>(i|f|s|qr)=(.*)</\\1>");

// ----------------------------------------------------------------------
// Convert the canvas into a PNG image.
class MyImageDump : public TImageDump
{
public:
  MyImageDump(TVirtualPad *pad, char *&data, int &size)
    {
      fImage = new TASImage;
      fType = 114;
      SetName(pad->GetName());
      SetBit(BIT(11));

      pad->Update();
      pad->Paint();

      fImage->GetImageBuffer(&data, &size, TImage::kPng);
    }
};

// ----------------------------------------------------------------------
VisDQMImageServer::VisDQMImageServer(const ServerInfo &info)
  : info_(info)
{
  // Scan the render plug-in catalogue and set up plug-ins.
  IgPluginManager::get()->initialise();
  DQMRenderPluginDB *db = DQMRenderPluginDB::get();
  for (DQMRenderPluginDB::Iterator i = db->begin(), e = db->end(); i != e; ++i)
  {
    plugins_.push_back(db->create(i));
    plugins_.back()->initialise(info.argc, info.argv);
  }
  applied_.resize (plugins_.size(), false);
}

VisDQMImageServer::~VisDQMImageServer(void)
{
  for (size_t i = 0, e = plugins_.size(); i != e; ++i)
    delete plugins_[i];
}

// Parse an integer parameter in an image spec.  If @ap starts with
// the parameter prefix @a name of length @a len, extracts the number
// that follows the prefix into @a value and advances @a p to the next
// character after the extracted value.  Returns true if the parameter
// was parsed, false otherwise.
bool
VisDQMImageServer::parseInt(const char *&p, const char *name, size_t len, int &value)
{
  if (! strncmp(p, name, len))
  {
    value = strtol(p+len, (char **) &p, 10);
    return true;
  }
  return false;
}

// Parse a double parameter in an image spec, like parseInt().
bool
VisDQMImageServer::parseDouble(const char *&p, const char *name, size_t len, double &value)
{
  if (! strncmp(p, name, len))
  {
    value = strtod(p+len, (char **) &p);
    return true;
  }
  return false;
}

// Parse a boolean choice in an image spec, like parseInt().
bool
VisDQMImageServer::parseBoolean(const char *&p, const char *name, size_t len, bool &value)
{
  if (! strncmp(p, name, len))
  {
    p += len;
    if (! strncmp(p, "yes", 3))
    {
      value = true;
      p += 3;
      return true;
    }
    else if (! strncmp(p, "no", 2))
    {
      value = false;
      p += 2;
      return true;
    }
  }
  return false;
}

// Parse an axis type parameter in an image spec, like parseInt().
bool
VisDQMImageServer::parseAxisType(const char *&p, const char *name, size_t len, std::string &value)
{
  if (! strncmp(p, name, len))
  {
    p += len;
    if (! strncmp(p, "lin", 3)
	|| ! strncmp(p, "log", 3))
    {
      value.clear();
      value.insert(value.end(), p, p+3);
      p += 3;
      return true;
    }
  }
  return false;
}

// Parse ROOT draw options parameter in an image spec, like parseInt().
bool
VisDQMImageServer::parseOption(const char *&p, const char *name, size_t len, std::string &value)
{
  static const char SAFE [] = { "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
				"abcdefghijklmnopqrstuvwxyz"
				"01234567890 "  };
  if (! strncmp(p, name, len))
  {
    p += len;
    len = 0;
    while (p[len] && strchr(SAFE, p[len]))
      ++len;
    value.insert(value.end(), p, p+len);
    p += len;
    return true;
  }
  return false;
}

// Build a parsed image specification for an object image.  Returns
// already parsed image data if it exists, otherwise breaks the data
// into its components and adds a new empty image into the object.
// The image specification is expected to be a series of "key=value"
// pairs separated by semicolons.
VisDQMImgInfo *
VisDQMImageServer::parseImageSpec(VisDQMObject &o, const std::string &spec, const char *&error)
{
  error = 0;

  // If the image already exists, return it.
  VisDQMImgMap::iterator pos = o.images.find (spec);
  if (pos != o.images.end ())
    return &pos->second;

  // Build new data by parsing the specification.
  VisDQMImgInfo &i = o.images[spec];
  i.version = ~(uint64_t)0;
  i.imgspec = spec;
  i.width = -1;
  i.height = -1;
  i.reference = false;
  i.xaxis.type = "lin";
  i.xaxis.min = NAN;
  i.xaxis.max = NAN;
  i.yaxis.type = "lin";
  i.yaxis.min = NAN;
  i.yaxis.max = NAN;
  i.zaxis.type = "lin";
  i.zaxis.min = NAN;
  i.zaxis.max = NAN;

  const char *p = spec.c_str();
  while (*p)
  {
    // Skip separators.
    while (*p == ';')
      ++p;

    // Handle various keywords.
    error = "unexpected image parameter";
    if (!    parseInt      (p, "w=",        2, i.width)
	&& ! parseInt      (p, "h=",        2, i.height)
	&& ! parseBoolean  (p, "withref=",  8, i.reference)
	&& ! parseDouble   (p, "xmin=",     5, i.xaxis.min)
	&& ! parseDouble   (p, "xmax=",     5, i.xaxis.max)
	&& ! parseAxisType (p, "xtype=",    6, i.xaxis.type)
	&& ! parseDouble   (p, "ymin=",     5, i.yaxis.min)
	&& ! parseDouble   (p, "ymax=",     5, i.yaxis.max)
	&& ! parseAxisType (p, "ytype=",    6, i.yaxis.type)
	&& ! parseDouble   (p, "zmin=",     5, i.zaxis.min)
	&& ! parseDouble   (p, "zmax=",     5, i.zaxis.max)
	&& ! parseAxisType (p, "ztype=",    6, i.zaxis.type)
	&& ! parseOption   (p, "drawopts=", 9, i.drawOptions))
    {
      o.images.erase(spec);
      return 0;
    }

    if (*p && *p != ';')
    {
      error = "unexpected characters in parameter value";
      o.images.erase(spec);
      return 0;
    }
  }

  if (i.width < 0 || i.height < 0)
  {
    error = "image dimensions missing, zero or negative";
    o.images.erase(spec);
    return 0;
  }

  return &i;
}

// ----------------------------------------------------------------------
// Render a ROOT object.
void
VisDQMImageServer::doRender(VisDQMObject &o, VisDQMImgInfo &i)
{
  // Clear the image data first.
  i.imgdata.clear ();

  // Record if the object is black-listed for special drawing.  Tell
  // the parent process which object we are now drawing so if we crash
  // here the parent will black-list this object.
  memset(info_.crashobj, 0, CRASH_NAME_LEN);
  strncpy(info_.crashobj, o.name.c_str(),
	  std::min (o.name.size(), CRASH_NAME_LEN-1));

  // Give me 30 seconds to render the histogram, then die.
  struct itimerval timer;
  timer.it_value.tv_sec = 30;
  timer.it_value.tv_usec = 0;
  timer.it_interval.tv_sec = 0;
  timer.it_interval.tv_usec = 0;
  setitimer(ITIMER_VIRTUAL, &timer, 0);

  // ROOT drawing.
  VisDQMRenderInfo ri = { i.drawOptions, (info_.blacklist.count(o.name) > 0) };
  logme()
    << "INFO: rendering '" << o.name << "' version " << o.version
    << " as '" << i.imgspec << (ri.blacklisted ? "', black-listed" : "'")
    << ", " << (o.reference ? "has" : "no") << " reference\n";

  // Prepare render defaults and the canvas.
  gStyle->Reset();
  gStyle->SetFillColor(10);
  gStyle->SetCanvasColor(10);
  gStyle->SetTitleFillColor(10);
  gStyle->SetTitleBorderSize(0);
  gStyle->SetOptStat(1111);
  gStyle->SetStatColor(10);
  gStyle->SetStatBorderSize(1);
  gROOT->ForceStyle();
  if (o.object)
    o.object->UseCurrentStyle();

  TCanvas c("", "", i.width+4, i.height+28);
  c.SetFillColor(TColor::GetColor(255, 255, 255));
  c.SetBorderMode(0);

  // Force histogram bounds reset now.  Plug-ins can change them.
  // We need this for the axis bounds code to work.
  if (TH1 *h = dynamic_cast<TH1 *>(o.object))
  {
    h->SetMinimum();
    h->SetMaximum();
  }

  // Invoke pre-draw hook on plug-ins that want to apply.
  for (size_t p = 0; p < plugins_.size(); ++p)
    if (! ri.blacklisted && plugins_[p] && plugins_[p]->applies (o, i))
    {
      plugins_[p]->preDraw(&c, o, i, ri);
      applied_[p] = true;
    }
    else
      applied_[p] = false;

  // Actually draw something.  If there's no object, inform user.
  double xmin = NAN, xmax = NAN;
  double ymin = NAN, ymax = NAN;
  double zmin = NAN, zmax = NAN;
  TObject *ob = 0;
  if (! o.object)
  {
    TText *t = new TText (.5, .5, "(Unavailable)");
    t->SetNDC(kTRUE);
    t->SetTextSize(0.10);
    t->SetTextAlign(22);
    t->SetTextColor(TColor::GetColor(32, 32, 178));
    t->Draw();
  }

  // It it exists but was black-listed, inform user.
  else if (ri.blacklisted)
  {
    TText *t = new TText (.5, .5, "(ROOT crashed)");
    t->SetNDC(kTRUE);
    t->SetTextSize(0.10);
    t->SetTextAlign(22);
    t->SetTextColor(TColor::GetColor(178, 32, 32));
    t->Draw();
  }

  // It's there and wasn't black-listed, paint it.
  else
  {
    ob = o.object;
    if (TObjString *os = dynamic_cast<TObjString *>(o.object))
    {
      TText *t;
      RegexpMatch m;
      if (s_rxmeval.match(os->GetName(), 0, 0, &m))
	t = new TText(.5, .5, m.matchString(os->GetName(), 3).c_str());
      else
	t = new TText(.5, .5, os->GetName());

      t->SetTextFont (63);
      t->SetTextAlign (22);
      t->SetTextSize (16);
      ob = t;
    }

#if 0
    // If we need to plot a reference and have one, prepare to
    // normalise the two.
    TH1F *th1f = dynamic_cast<TH1F *>(ob);
    if (i.reference && o.reference && th1f)
    {
      th1f->SetMinimum(0.);
      double max = th1f->GetMaximum();
      double min = th1f->GetMinimum();
      th1f->SetMaximum(max*1.2);
      if (max == 0)
        th1f->SetMaximum(1.1);
      th1f->SetMinimum(min*1.2);
      if (max == 0 && min == 0)
        th1f->SetMinimum(-1.1);
      th1f->SetLineWidth(2);
    }
#endif

    // If we have an object that has axes, apply the requested params.
    // Set only the bounds that were specified, and leave the rest to
    // the natural range.  Unset bounds have NaN as their value.
    if (TH1 *h = dynamic_cast<TH1 *>(ob))
    {
      if (TAxis *a = h->GetXaxis())
      {
	if (! (isnan(i.xaxis.min) && isnan(i.xaxis.max)))
        {
	  xmin = a->GetXmin();
	  xmax = a->GetXmax();
	  a->SetRangeUser(isnan(i.xaxis.min) ? xmin : i.xaxis.min,
			  isnan(i.xaxis.max) ? xmax : i.xaxis.max);
        }

        if (i.xaxis.type == "log")
	  c.SetLogx();
      }

      if (TAxis *a = h->GetYaxis())
      {
	if (! (isnan(i.yaxis.min) && isnan(i.yaxis.max)))
        {
	  ymin = a->GetXmin();
	  ymax = a->GetXmax();
	  a->SetRangeUser(isnan(i.yaxis.min) ? ymin : i.yaxis.min,
			  isnan(i.yaxis.max) ? ymax : i.yaxis.max);
        }

        if (i.yaxis.type == "log")
          c.SetLogy();
      }

      if (TAxis *a = h->GetZaxis())
      {
	if (! (isnan(i.zaxis.min) && isnan(i.zaxis.max)))
        {
	  zmin = a->GetXmin();
	  zmax = a->GetXmax();
	  a->SetRangeUser(isnan(i.zaxis.min) ? zmin : i.zaxis.min,
			  isnan(i.zaxis.max) ? zmax : i.zaxis.max);
        }

        if (i.zaxis.type == "log")
          c.SetLogz();
      }
    }

    ob->Draw(ri.drawOptions.c_str());

    // Now maybe draw the reference.
    if (i.reference && o.reference)
    {
      if (TH1F *ref = dynamic_cast<TH1F *>(o.reference))
      {
        double norm = 1.;
        if (TH1F *th1f = dynamic_cast<TH1F *>(ob))
          norm = th1f->GetEntries();
        ref->SetLineColor(13);
        ref->SetLineWidth(0);
        ref->SetFillColor(13);
        ref->SetFillStyle(3002);
        ref->GetListOfFunctions()->Delete();
        if (norm)
          ref->DrawNormalized("samehist",norm);
        else
          ref->Draw("samehist");
      }
      if (TProfile *ref = dynamic_cast<TProfile *>(o.reference))
      {
        ref->SetLineColor(13);
        ref->SetLineWidth(0);
        ref->GetListOfFunctions()->Delete();
        ref->Draw("samehist");
      }
    }
  }

  // Invoke post-draw hook on plug-ins that applied.
  for (size_t p = 0; p < plugins_.size(); ++p)
    if (applied_[p])
      plugins_[p]->postDraw(&c, o, i);

  char		*data = 0;
  int		size = 0;
  MyImageDump	idump (&c, data, size);

  // Grab image data and remember what version this object was.
  i.imgdata.insert (i.imgdata.end(),
		    (unsigned char *) data,
		    (unsigned char *) data + size);
  i.version = o.version;
  free (data);

  // Restore axes to avoid permanent changes.
  if (TH1 *h = dynamic_cast<TH1 *>(ob))
  {
    TAxis *a;
    if ((a = h->GetXaxis()) && ! (isnan(xmin) && isnan(xmax)))
      a->SetRangeUser(xmin, xmax);

    if ((a = h->GetYaxis()) && ! (isnan(ymin) && isnan(ymax)))
      a->SetRangeUser(ymin, ymax);

    if ((a = h->GetZaxis()) && ! (isnan(zmin) && isnan(zmax)))
      a->SetRangeUser(zmin, zmax);

    h->SetMinimum();
    h->SetMaximum();
  }

  // Let timer go now, we made it.
  timer.it_value.tv_sec = 0;
  timer.it_value.tv_usec = 0;
  timer.it_interval.tv_sec = 0;
  timer.it_interval.tv_usec = 0;
  setitimer(ITIMER_VIRTUAL, &timer, 0);
}
